# Gcode_Exporter.rb - Exports gcode .gcode file from SketchUp
#
# Copyright (C) 2013 David Fishelzon
#
# License: Apache License, Version 2.0
#
# G-Code Exporter
#
require 'sketchup'

module GcodeExporterModule  # Namespace for THIS plugin

  class GcodeExporter
    unless file_loaded?(__FILE__) 
      Gcode_Exporter_File_Heder  = ";Gcode Exporter File Heder"
      Gcode_Exporter_Layer_Heder = "; Gcode Exporter Layer %s Heder "
      Gcode_Exporter_Layer_Futer = "; Gcode Exporter Layer %s Futer "
      Gcode_Exporter_File_Futer  = ";Gcode Exporter File Futer"
    end #file_loaded?(__FILE__) 

    #
    # exportLayerGcode exports the gcode in a single layer
    #
    def exportLayerGcode startPoint
      currPoint = startPoint
      begin
        # Find the segment at currPoint
        @ptr = 0
        while @ptr < @entsInLayer.length
          if currPoint.vector_to(@entsInLayer[@ptr][0]).length.to_f < 0.01.mm
            # while - Found the next segment!
            currPoint = @entsInLayer.delete_at(@ptr)
            x = "X%0.3f" % (currPoint[1][0] / 1.mm)
            y = "Y%0.3f" % (currPoint[1][1] / 1.mm)
            z = "Z%0.3f" % (currPoint[1][2] / 1.mm)
            if currPoint[2] 
              # component is a filament 
              l = "E%0.3f" % (currPoint[0].vector_to(currPoint[1]).length.to_f / 1.mm)
            else
              # component is a Head Move
              l = "E0"
            end # if component a filament or a Head move
            @gcodeFile.puts "G1 #{x} #{y} #{z} #{l} "
            currPoint = currPoint[1]
            @ptr = @entsInLayer.length + 1 # stop looking - move to end +1
          else # try next segment
            @ptr += 1
          end # if
        end # while @ptr < @entsInLayer.length
      end until @entsInLayer.length == 0 or @ptr == @entsInLayer.length
      return currPoint
    end #exportLayerGcode

    #
    # exportGcode will sort all connected Gcode components by layers
    # 
    def exportGcode
      @gcodeFile.puts Gcode_Exporter_File_Heder

      # Get Layar list and an array for each layer's components
      @layersList    = []
      @export_ents.map {|ent| 
        a_headMove   = nil
        a_filament   = @filamentComp.instances.include? ent
        edge         = ent.entities[12] if a_filament
        unless a_filament # chack maybe ent is a head move
          a_headMove = @HeadMoveComp.instances.include? ent 
          edge       = ent.entities[0] if a_headMove
        end # unless a_filament
        if a_filament or a_headMove
          lay     = ent.layer.name
          layEval = "@lay_" + lay.split.join.downcase
          unless @layersList.include?( lay )
            @layersList << lay
            eval "#{layEval} = [] "
          end #  @layersList.include?( lay )
          trnsf = ent.transformation
          strPo = edge.start.position.transform!(trnsf)
          endPo = edge.end.position.transform!(trnsf)
          eval "#{layEval}  << [strPo.to_a, endPo.to_a, a_filament] "
        end # if ent.is_a Component
      } # @export_ents.map
      @layersList.sort!

      point = Geom::Point3d.new(0, 0, 0)
      # export all layers
      @layersList.each  {|lay|
        layEval = "@lay_" + lay.split.join.downcase
        eval "@entsInLayer = #{layEval} "
        if @entsInLayer.length > 0
          @gcodeFile.puts Gcode_Exporter_Layer_Heder % lay
          point = self.exportLayerGcode point
          @gcodeFile.puts Gcode_Exporter_Layer_Futer % lay
        end # if
        if @entsInLayer.length > 0
          puts " #{@entsInLayer.length} Components are not connected at layer #{lay} "
        end # if
        Sketchup.status_text="finished exporting layer #{lay} "
      } # @layersList.each
      
      @gcodeFile.puts Gcode_Exporter_File_Futer
    end # def exportGcode

    def exportGcodeUI 
      @model        = Sketchup.active_model
      @filamentComp = @model.definitions[ "filament" ]
      return nil   if not @filamentComp
      @HeadMoveComp = @model.definitions[ "HeadMove" ]
      return nil   if not @HeadMoveComp 

      # Start Undo
      if (Sketchup.version.to_i < 7)
        @model.start_operation("Export G-Code")
      else
        @model.start_operation("Export G-Code",true)
      end

      if @model.selection.empty?
        @export_ents = @model.entities
      else
        @export_ents = @model.selection
      end

      if ( @export_ents.length > 0 )
        # Get exported file name and export.
        @out_name = File.basename(@model.path).split(".")[0]
        @out_name = "untitled" if not @out_name
        @out_name = UI.savepanel( "Gcode file location", "", "#{@out_name}.gcode" )
        if @out_name
          @gcodeFile = File.new( @out_name , "w" )  
          puts "~~~~Starting~~~~~"
          Sketchup.status_text="Gcode export Stated"
          self.exportGcode()

          puts "~~~~FINISH~~~~~"
          Sketchup.status_text=""
          @gcodeFile.close
          UI.messagebox( " Gcode file exported #{@out_name}" )
        end # @out_name exist
      end # if length > 0
     
      @model.commit_operation # End Undo
    end # def exportGcodeUI 

  end # class GcodeExporter

  # Register Menus in the UI
  unless file_loaded?(__FILE__) 

    UI.menu('Plugins').add_item("Fishelzon's G-Code Exporter...")     { 
      GcodeExporter.new.exportGcodeUI 
    } # UI.menu
=begin
    UI.menu('Plugins').add_item("Fishelzon's G-Code Exporter Reload") { 
      if load( __FILE__ ) 
        Sketchup.send_action 'showRubyPanel:'
      end #if load( __FILE__ ) 
    } # UI.menu Reload
=end
  end # unless file_loaded?
  file_loaded(__FILE__)

end # module GcodeExporterModule
